import typing

import sqlalchemy as sa

from . import interface

T = typing.TypeVar("T")


def _make_converter(
    converter: typing.Callable[[interface.FieldType], interface.FieldType],
):
    def _converter(
        comparator: interface.Comparator[T],
    ) -> interface.Comparator[T]:
        def _comp(field: interface.FieldType, target: T) -> interface.Comparison:
            return comparator(converter(field), target)

        return _comp

    return _converter


as_date = _make_converter(sa.func.date)
as_time = _make_converter(sa.func.time)
as_lower = _make_converter(sa.func.lower)
as_upper = _make_converter(sa.func.upper)

__all__ = ["as_date", "as_time", "as_lower", "as_upper"]



import typing
from datetime import date
from datetime import time

import sqlalchemy as sa
from sqlalchemy.sql import Delete
from sqlalchemy.sql import Select
from sqlalchemy.sql import Update
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql.functions import Function

from gyver.database.entity import AbstractEntity
from gyver.database.typedef import ClauseType

ExecutableType = typing.Union[Select, Update, Delete]
ExecutableT = typing.TypeVar("ExecutableT", bound=ExecutableType)
Sortable = typing.Union[int, float, date, time]
Comparison = typing.Union[ColumnElement[sa.Boolean], BooleanClauseList, Function]
FieldType = typing.Union[ColumnElement, sa.Column]
T = typing.TypeVar("T", contravariant=True)
Mapper = typing.Union[sa.Table, type[AbstractEntity]]


class Comparator(typing.Protocol[T]):
    def __call__(self, field: FieldType, target: T) -> typing.Any:
        pass


class Clause(typing.Protocol):
    type_: typing.ClassVar[ClauseType]


class BindClause(Clause, typing.Protocol):
    type_: typing.Literal[ClauseType.BIND]

    def bind(self, mapper: Mapper) -> Comparison:
        ...


class ApplyClause(Clause, typing.Protocol[ExecutableT]):
    type_: typing.Literal[ClauseType.APPLY]

    def apply(self, query: ExecutableT) -> ExecutableT:
        ...



from collections import defaultdict
from typing import Mapping
from typing import Optional
from urllib.parse import parse_qs
from urllib.parse import quote

from gyver.url.encode import Encodable


class Query(Encodable):
    def __init__(self, querystr: str) -> None:
        self.params: defaultdict[str, list[str]] = defaultdict(list)
        self.params.update(parse_qs(querystr, keep_blank_values=True))

    def encode(self):
        return "&".join(
            "=".join((quote(key, safe=""), quote(item or "", safe="")))
            for key, value in self.params.items()
            for item in value
        )

    def omit_empty_equal(self):
        return "&".join(
            "=".join(map(quote, (key, item))) if item else quote(key)
            for key, value in self.params.items()
            for item in value
        )

    def add(self, args: Optional[Mapping[str, str]] = None, /, **params: str):
        query = {**(args or {}), **params}
        for key, value in query.items():
            self.params[key].append(value)
        return self

    def set(self, args: Optional[Mapping[str, str]] = None, /, **params: str):
        query = {**(args or {}), **params}
        self.params.clear()
        self.add(query)
        return self

    def __setitem__(self, key: str, value: str):
        self.add({key: value})

    def __getitem__(self, key: str) -> list[str]:
        return self.params[key]

    def remove(self, *keys: str):
        for key in keys:
            self.params.pop(key, None)
        return self

    def sort(self):
        self.params = defaultdict(list, sorted(self.params.items()))
        return self



from typing import Optional
from urllib.parse import quote

from gyver.url.encode import Encodable
from gyver.url.utils import utf8


class Netloc(Encodable):
    def __init__(self, netloc: str) -> None:
        self.username: Optional[str] = None
        self.password: Optional[str] = None
        self.load(netloc)

    def load(self, netloc: str):
        userinfo, _, host = netloc.partition("@")
        port = None
        if host:
            username, _, password = userinfo.partition(":")
            self.username = username
            self.password = password
        else:
            host = userinfo
        if ":" in host:
            host, _, port = host.partition(":")
        self.host = host
        port = int(port) if port else None
        self.port = port

    def encode(self):
        netloc = ""
        if self.username:
            netloc = f"{quote(utf8(self.username))}"
            if self.password:
                netloc += f":{quote(utf8(self.password))}"
            netloc += "@"
        netloc += self.host
        if self.port:
            netloc += f":{self.port}"
        return netloc

    def parse(self, netloc: str):
        """
        Parse the given netloc string and populate the properties of the Netloc object.
        """
        if "@" in netloc:
            userinfo, host = netloc.split("@", 1)
            if ":" in userinfo:
                self.username, self.password = userinfo.split(":", 1)
            else:
                self.username = userinfo
            self.host = host
        elif ":" in netloc:
            self.host, port = netloc.split(":", 1)
            if port:
                self.port = int(port)
        else:
            self.host = netloc

    def set(
        self,
        host: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        port: Optional[int] = None,
    ):
        self.host = host or self.host
        self.username = username or self.username
        self.password = password or self.password
        self.port = port or self.port
        return self

    def merge(self, netloc: "Netloc") -> "Netloc":
        host = netloc.host or self.host
        username = netloc.username or self.username
        password = netloc.password or self.password
        port = netloc.port or self.port
        return self.from_args(host, username, password, port or None)

    @classmethod
    def from_args(
        cls,
        host: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        port: Optional[int] = None,
    ):
        netloc = cls("")
        return netloc.set(host, username, password, port)



import typing

import sqlalchemy as sa

from gyver.database.query.null import NullBind

from .interface import BindClause
from .interface import Comparator
from .interface import Comparison
from .interface import FieldType
from .interface import Sortable


def always_true(field: FieldType, target: typing.Any) -> Comparison:
    del field, target
    return sa.true()


def equals(field: FieldType, target: typing.Any) -> Comparison:
    return field == target


def not_equals(field: FieldType, target: typing.Any) -> Comparison:
    return field != target


def greater(field: FieldType, target: Sortable) -> Comparison:
    return field > target


def greater_equals(field: FieldType, target: Sortable) -> Comparison:
    return field >= target


def lesser(field: FieldType, target: Sortable) -> Comparison:
    return field < target


def lesser_equals(field: FieldType, target: Sortable) -> Comparison:
    return field <= target


def between(field: FieldType, target: tuple[Sortable, Sortable]) -> Comparison:
    left, right = target
    return field.between(left, right)


def range(field: FieldType, target: tuple[Sortable, Sortable]) -> Comparison:
    left, right = target
    return sa.and_(greater_equals(field, left), (lesser(field, right)))


def like(field: FieldType, target: str) -> Comparison:
    return field.like(f"%{target}%")


def rlike(field: FieldType, target: str) -> Comparison:
    return field.like(f"{target}%")


def llike(field: FieldType, target: str) -> Comparison:
    return field.like(f"%{target}")


def insensitive_like(
    opt: typing.Literal["like", "rlike", "llike"] = "like"
) -> Comparator[str]:
    fmt = {"like": "%{target}%", "rlike": "%{target}", "llike": "{target}%"}[
        opt
    ]

    def comparator(field: FieldType, target: str):
        return field.ilike(fmt.format(target=target))

    return comparator


def isnull(field: FieldType, target: bool) -> Comparison:
    return field.is_(None) if target else field.is_not(None)


def includes(field: FieldType, target: typing.Sequence) -> Comparison:
    return field.in_(list(target))


def excludes(field: FieldType, target: typing.Sequence) -> Comparison:
    return field.not_in(list(target))


def json_contains(field: FieldType, target: typing.Any) -> Comparison:
    return sa.func.json_contains(field, f'"{target}"')


def json_empty(field: FieldType, target: typing.Any) -> Comparison:
    func_length = sa.func.json_length(field)
    return func_length == 0 if target else func_length != 0


def make_relation_check(clause: BindClause) -> Comparator:
    def _relation_exists(field: FieldType, target: bool) -> Comparison:
        comp = clause.bind(field.class_)
        func = (
            field.has
            if field.property.direction.name.lower()
            not in ("onetomany", "manytomany")
            else field.any
        )
        result = func() if str(comp) == str(sa.true()) else func(comp)
        return result if target else ~result

    return _relation_exists



from contextlib import suppress
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any, cast
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import get_args
from typing import get_origin

from pydantic import BaseModel

from gyver.attrs import mark_factory
from gyver.config.config import MISSING
from gyver.config.config import Config
from gyver.config.utils import boolean_cast
from gyver.exc import MissingName
from gyver.utils import finder
from gyver.utils import json
from gyver.utils import panic
from gyver.utils.strings import make_lex_separator

from .attrs import AttrsResolverStrategy
from .dataclass import DataclassResolverStrategy
from .gattrs import GyverAttrsResolverStrategy
from .interface import FieldResolverStrategy
from .mark import is_config
from .pydantic import PydanticResolverStrategy

_default_config = Config()

T = TypeVar("T")


def _try_each(*names: str, default: Any, cast: Any, config: Config):
    for name in names:
        with suppress(MissingName):
            return config(name, cast)
    if default is not MISSING:
        return default
    raise panic(
        MissingName, f"{', '.join(names)} not found and no default was given"
    )


def _resolve_cast(outer_type: type):
    _sequences = (list, tuple, set)
    origin = get_origin(outer_type)
    if outer_type is bool:
        return boolean_cast
    if origin is None:
        return (
            make_lex_separator(outer_type)
            if outer_type in _sequences
            else outer_type
        )
    if (origin := get_origin(outer_type)) in _sequences:
        args = get_args(outer_type)
        cast = args[0] if args else str
        return make_lex_separator(origin, cast)  # type: ignore
    return _loads if dict in (origin, outer_type) else origin


def _loads(val: Any) -> Any:
    return json.loads(val) if isinstance(val, str) else val


class AdapterConfigFactory:
    def __init__(self, config: Config = _default_config) -> None:
        self._config = config

    def get_strategy_class(
        self, config_class: type
    ) -> type[FieldResolverStrategy]:
        if hasattr(config_class, "__gyver_attrs__"):
            return GyverAttrsResolverStrategy
        elif is_dataclass(config_class):
            return DataclassResolverStrategy
        elif issubclass(config_class, BaseModel):
            return PydanticResolverStrategy
        elif hasattr(config_class, "__attrs_attrs__"):
            return AttrsResolverStrategy
        raise ValueError("Unknown class definition")

    def load(
        self,
        model_cls: type[T],
        __prefix__: str = "",
        *,
        presets: Optional[Mapping[str, Any]] = None,
        **defaults: Any,
    ) -> T:
        presets = presets or {}
        strategy_class = self.get_strategy_class(model_cls)
        resolvers = tuple(
            resolver
            for field in strategy_class.iterfield(model_cls)
            if (resolver := strategy_class(field)).init_name() not in presets
        )
        result = {
            resolver.init_name(): self._get_value(
                model_cls, resolver, __prefix__, defaults
            )
            for resolver in resolvers
        }
        return model_cls(**result | presets)

    def maker(
        self,
        model_cls: type[T],
        __prefix__: str = "",
        *,
        presets: Optional[Mapping[str, Any]] = None,
        **defaults: Any,
    ) -> Callable[[], T]:
        @mark_factory
        def load():
            return self.load(
                model_cls, __prefix__, presets=presets, **defaults
            )

        return load

    def _get_value(
        self,
        model_cls: type,
        resolver: FieldResolverStrategy,
        prefix: str,
        defaults: Mapping[str, Any],
    ):
        names = self.resolve_names(model_cls, resolver, prefix)
        default = next(
            (result for name in names if (result := defaults.get(name))),
            resolver.default(),
        )
        cast = _resolve_cast(resolver.cast())
        return _try_each(
            *names, default=default, cast=cast, config=self._config
        )

    def resolve_names(
        self, model_cls: type, resolver: FieldResolverStrategy, prefix: str
    ) -> Sequence[str]:
        names = resolver.names()
        prefix = prefix or getattr(model_cls, "__prefix__", "")
        prefix = prefix.removesuffix("_")
        without_prefix = cast(
            Sequence[str], getattr(model_cls, "__without_prefix__", ())
        )
        if not prefix or set(names).intersection(without_prefix):
            return (*names, *map(str.upper, names))

        processed_names = []
        for name in names:
            name.lstrip("_")
            processed = f"{prefix}_{name}".lower()
            processed_names.extend((processed, processed.upper()))

        return tuple(processed_names)

    @classmethod
    def resolve_confignames(
        cls,
        root: Path,
    ) -> dict[type, tuple[Sequence[str], ...]]:
        """
        The resolve_confignames function resolves the names of all environment
        variables required by the configs in any class marked as config.
        It returns them in a dict of {class: (configs)}.

        :param cls: Call the class_validator function, which is used to validate that a
        class has all of the required attributes
        :param root: Path: Specify the root directory of the project
        :return: A dictionary of {class: (configs)}
        """
        builder = (
            finder.FinderBuilder().add_validator(is_config).from_path(root)
        )
        output = builder.find()
        tempself = cls()
        return {
            cfg: tuple(
                tempself.resolve_names(cfg, field, "")
                for field in cfg.__fields__.values()
            )
            for cfg in output.values()
        }



from collections import defaultdict
from typing import Mapping
from typing import Optional
from urllib.parse import urlparse
from urllib.parse import urlunsplit

from gyver.url.encode import Encodable
from gyver.url.fragment import Fragment
from gyver.url.netloc import Netloc
from gyver.url.path import Path
from gyver.url.query import Query


class URL(Encodable):
    def __init__(self, val: str) -> None:
        self.load(val)

    def load(self, val: str):
        parsed = urlparse(val)
        (
            self.scheme,
            netloc,
            path,
            _,
            query,
            fragment,
        ) = parsed
        self.query = Query(query)
        self.path = Path(path)
        self.port = parsed.port
        self.fragment = Fragment(fragment)
        self.netloc = Netloc(netloc)

    def encode(self, omit_empty_equal: bool = True):
        resolved_query = (
            self.query.encode() if omit_empty_equal else self.query.omit_empty_equal()
        )
        return urlunsplit(
            (
                self.scheme,
                self.netloc.encode(),
                self.path.encode(),
                resolved_query,
                self.fragment.encode(),
            )
        )

    def __repr__(self) -> str:
        return object.__repr__(self)

    def add(
        self,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_args: Optional[Netloc] = None,
    ):
        if queryasdict:
            self.query.add(queryasdict)
        if path:
            self.path.add(path)
        if query:
            self.query.add(query)
        if fragment:
            self.fragment.set(fragment)
        if netloc:
            self.netloc.set(netloc)
        if netloc_args:
            self.netloc = self.netloc.merge(netloc_args)
        return self

    def set(
        self,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_args: Optional[Netloc] = None,
    ):
        if queryasdict:
            self.query.set(queryasdict)
        if path:
            self.path.set(path)
        if query:
            self.query.set(query)
        if fragment:
            self.fragment.set(fragment)
        if netloc:
            self.netloc.set(netloc)
        if netloc_args:
            self.netloc = netloc_args
        return self

    def copy(self):
        new_url = URL("")
        new_url.scheme = self.scheme
        new_url.fragment.fragment_str = self.fragment.fragment_str
        new_url.path.segments = self.path.segments.copy()
        new_url.query.params = defaultdict(
            list,
            ((key, list(value)) for key, value in self.query.params.items()),
        )
        new_url.add(netloc_args=self.netloc)
        return new_url



import enum
import typing
from dataclasses import dataclass

from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql import Select


class OrderDirection(enum.Enum):
    ASC = "asc"
    DESC = "desc"


@dataclass(frozen=True)
class OrderBy:
    __slots__ = ("field", "direction")

    field: typing.Optional[str]
    direction: OrderDirection

    @property
    def _should_apply(self):
        return self.field is not None

    @classmethod
    def none(cls):
        return cls(field=None, direction=OrderDirection.ASC)

    @classmethod
    def asc(cls, field: str):
        return cls(field, OrderDirection.ASC)

    @classmethod
    def desc(cls, field: str):
        return cls(field, OrderDirection.DESC)

    def apply(self, query: Select) -> Select:
        return (
            query.order_by(self._apply_order(self._find_column(query)))
            if self._should_apply
            else query
        )

    def _find_column(self, query: Select) -> ColumnElement:
        try:
            return next(col for col in query.selected_columns if col.key == self.field)
        except StopIteration:
            raise ValueError(f"Field {self.field} does not exist in query") from None

    def _apply_order(self, col: ColumnElement):
        return col.asc() if self.direction is OrderDirection.ASC else col.desc()



from .asyncio import AsyncConnectionAdapter
from .asyncio import AsyncSaContext
from .sync import SaAdapter
from .sync import SaContext

__all__ = ["AsyncSaContext", "AsyncConnectionAdapter", "SaAdapter", "SaContext"]



from .asyncio import AsyncCacheProvider
from .config import CacheConfig
from .interface import AsyncCacheInterface
from .interface import CacheInterface
from .sync import CacheProvider
from .utils import make_uri

__all__ = [
    "AsyncCacheProvider",
    "CacheInterface",
    "AsyncCacheInterface",
    "CacheConfig",
    "make_uri",
    "CacheProvider",
]



from typing import Literal
from typing import Optional
from typing import overload

from gyver.utils import lazyfield

from .config import CacheConfig
from .interface import CacheInterface
from .mapper import CacheMap
from .mock import MockCache
from .redis import RedisWrapper


class CacheProvider:
    @overload
    def __init__(self, *, config: CacheConfig):
        ...

    @overload
    def __init__(self, *, test: Literal[True]):
        ...

    def __init__(
        self, *, config: Optional[CacheConfig] = None, test: bool = False
    ) -> None:
        self._config = config
        self._test = test

    @staticmethod
    def _make_concrete(
        config: Optional[CacheConfig], test: bool
    ) -> CacheInterface:
        if test:
            return MockCache()
        return RedisWrapper() if config is None else RedisWrapper(config)

    @lazyfield
    def interface(self):
        return self._make_concrete(self._config, self._test)

    def mapper(self, name: str) -> CacheMap:
        return CacheMap(self.interface, name)



import typing
from functools import partial
from types import FunctionType

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from gyver.database.metadata import metadata as default_metadata
from gyver.url import URL
from gyver.url.netloc import Netloc
from gyver.utils import cache

from . import drivers
from .config import DatabaseConfig


@cache
def make_uri(config: DatabaseConfig, sync: bool = False) -> str:
    """Construct a database URI from a database configuration .

    Args:
        config (DatabaseConfig): configuration object containing information about the
        database such as dialect, host, username, password, etc.
        sync (bool, optional): If uses async or sync drivers. Defaults to False.

    Returns:
        str: database URI in the format
        'dialect_name(+driver)?://username:password@host:port/dbname'
        if not only host else 'dialect_name(+driver)?://host'
    """

    url = URL("").set(netloc=config.host)
    url.scheme = f"{drivers.build_dialect_scheme(config.dialect, sync)}"

    if config.dialect.only_host:
        return url.encode()

    return url.add(
        path=config.name,
        netloc_args=Netloc("").set(
            username=config.user,
            password=config.password,
            port=config.effective_port,
        ),
    ).encode()


EntityT = typing.TypeVar("EntityT")
CallableT = typing.TypeVar("CallableT", bound=typing.Callable)


@typing.overload
def make_relation(
    relation: typing.Union[str, type[EntityT]],
    *,
    relation_name: str = "",
    back_populates: typing.Optional[str] = None,
    secondary: typing.Union[sa.Table, type, None] = None,
    foreign_keys: typing.Optional[list[typing.Any]] = None,
    lazy: str = "selectin",
    use_list: typing.Literal[True],
) -> list[EntityT]:
    ...


@typing.overload
def make_relation(
    relation: typing.Union[str, type[EntityT]],
    *,
    relation_name: str = "",
    back_populates: typing.Optional[str] = None,
    secondary: typing.Union[sa.Table, type, None] = None,
    foreign_keys: typing.Optional[list[typing.Any]] = None,
    lazy: str = "selectin",
    use_list: typing.Literal[False] = False,
) -> EntityT:
    ...


@typing.overload
def make_relation(
    relation: typing.Callable[[], type[EntityT]],
    *,
    relation_name: str = "",
    back_populates: typing.Optional[str] = None,
    secondary: typing.Union[sa.Table, type, None] = None,
    foreign_keys: typing.Optional[list[typing.Any]] = None,
    lazy: str = "selectin",
    use_list: typing.Literal[False] = False,
) -> EntityT:
    ...


@typing.overload
def make_relation(
    relation: typing.Callable[[], type[EntityT]],
    *,
    relation_name: str,
    back_populates: typing.Optional[str] = None,
    secondary: typing.Union[sa.Table, type, None] = None,
    foreign_keys: typing.Optional[list[typing.Any]] = None,
    lazy: str = "selectin",
    use_list: typing.Literal[True],  # pylint: disable=unused-argument
) -> list[EntityT]:
    ...


def make_relation(
    relation: typing.Union[str, type[EntityT], typing.Callable[[], type[EntityT]]],
    *,
    relation_name: str = "",
    back_populates: typing.Optional[str] = None,
    secondary: typing.Union[sa.Table, type, None] = None,
    foreign_keys: typing.Optional[list[typing.Any]] = None,
    lazy: str = "selectin",
    use_list: bool = False,  # noqa
) -> typing.Union[EntityT, typing.Sequence[EntityT]]:
    rel_func = partial(
        relationship,
        back_populates=back_populates,
        lazy=lazy,
        secondary=secondary,
        foreign_keys=foreign_keys,
    )
    if isinstance(relation, str):
        return typing.cast(typing.Any, rel_func(argument=relation))
    if isinstance(relation, FunctionType):
        return typing.cast(typing.Any, rel_func(argument=relation_name))
    return typing.cast(typing.Any, rel_func(argument=relation.__qualname__))


def create_relation_table(table_name: str, *entities: str):
    return sa.Table(
        table_name,
        default_metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        *[
            sa.Column(f"{entity}_id", sa.Integer, sa.ForeignKey(f"{entity}.id"))
            for entity in entities
        ],
    )



from enum import Enum


class Driver(str, Enum):
    MYSQL = "mysql"
    POSTGRES = "postgres"
    SQLITE = "sqlite"
    MARIADB = "mariadb"
    CUSTOM = "custom"


class OrderDirection(str, Enum):
    ASC = "asc"
    DESC = "desc"
    DISABLED = "disabled"


class ClauseType(str, Enum):
    BIND = "bind"
    APPLY = "apply"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gyver',
 'gyver.cache',
 'gyver.config',
 'gyver.config.adapter',
 'gyver.context',
 'gyver.context.atomic_',
 'gyver.context.interfaces',
 'gyver.crypto',
 'gyver.database',
 'gyver.database.context',
 'gyver.database.drivers',
 'gyver.database.query',
 'gyver.url',
 'gyver.utils']

package_data = \
{'': ['*']}

install_requires = \
['cryptography>=38.0.4,<39.0.0',
 'gyver-attrs>=0.1.4,<0.2.0',
 'orjson>=3.8.1,<4.0.0',
 'pydantic[email]>=1.10.2,<2.0.0',
 'typing-extensions>=4.4.0,<5.0.0']

extras_require = \
{':sys_platform != "linux"': ['tzdata>=2022.6,<2023.0'],
 'cache': ['redis>=4.4.0,<5.0.0'],
 'db-mariadb': ['pymysql>=1.0.2,<2.0.0', 'asyncmy>=0.2.5,<0.3.0'],
 'db-mysql': ['pymysql>=1.0.2,<2.0.0', 'asyncmy>=0.2.5,<0.3.0'],
 'db-pg': ['psycopg2>=2.9.5,<3.0.0', 'asyncpg>=0.27.0,<0.28.0'],
 'db-sqlite': ['aiosqlite>=0.18.0,<0.19.0'],
 'sqlalchemy': ['sqlalchemy>=2.0.4,<3.0.0']}

setup_kwargs = {
    'name': 'gyver',
    'version': '2.0.0',
    'description': 'Toolbox for web development',
    'long_description': '# Gyver\n\n> Simple toolbox for python development to skip code boilerplate.\n\n[**Documentation**](https://guscardvs.github.io/gyver/)\n\n[**Source Code**](https://github.com/guscardvs/gyver)\n\n## Authors\n\n> [@guscardvs](https://github.com/guscardvs)\n\n## Requirements\n\n* Python 3.9+\n\n## Required\n\n* [OrJSON](https://github.com/ijl/orjson) for json parsing.\n* [Pydantic](https://docs.pydantic.dev) for data handling.\n* [Typing Extensions](https://github.com/python/typing_extensions) for compatibility.\n* [Cryptography](https://cryptography.io) to handle encryption.\n\n## Optional\n\nTo use the database parts:\n* **Mysql/MariaDB:** AsyncMy, PyMySQL (use db-mysql or db-mariadb extras)\n* **Postgres:** AsyncPG, Psycopg2\n* **SQLite:** aiosqlite\n* **Redis:** redis\n* And **SQLAlchemy**\n\n## Installation\n\n```console\n$ pip install gyver\n```\n\n## Roadmap\n\n> Migrate from pydantic to dataclass/attrs for a lighter version of the library\n\n##  License\n\nThis project is licensed under the terms of the MIT license.',
    'author': 'Gustavo Correa',
    'author_email': 'self.gustavocorrea@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/guscardvs/gyver',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa

from gyver.config import AdapterConfigFactory
from gyver.config import as_config
from gyver.utils import lazyfield


@as_config
class RSACryptoConfig:
    __prefix__ = "crypto"

    private_key: str
    public_key: str


class RSACryptoProvider:
    def __init__(self, config: Optional[RSACryptoConfig] = None) -> None:
        self._config = config or AdapterConfigFactory().load(RSACryptoConfig)

    @property
    def config(self):
        return self._config

    @lazyfield
    def _private_key(self) -> rsa.RSAPrivateKey:
        return serialization.load_pem_private_key(  # type: ignore
            self._config.private_key.encode(),
            password=None,
            backend=default_backend(),
        )

    @lazyfield
    def _public_key(self) -> rsa.RSAPublicKey:
        return serialization.load_pem_public_key(  # type:ignore
            self._config.public_key.encode(), backend=default_backend()
        )

    def encrypt(self, secret: str) -> str:
        ciphertext = self._public_key.encrypt(
            secret.encode(),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        return ciphertext.hex()

    def decrypt(self, ciphertext: str) -> str:
        encoded = bytes.fromhex(ciphertext)
        return self._private_key.decrypt(
            encoded,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        ).decode()



import typing

import sqlalchemy as sa
import sqlalchemy.engine as sa_engine
import sqlalchemy.orm as sa_orm

from gyver import context
from gyver.utils import lazyfield

TransactionOption = typing.Optional[typing.Literal["open", "begin"]]

SaContext = context.Context[sa_engine.Connection]
SessionContext = context.Context[sa_orm.Session]


class SaAdapter(context.Adapter[sa_engine.Connection]):
    @typing.overload
    def __init__(
        self,
        *,
        uri: str,
        engine: None = None,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self,
        *,
        uri: None = None,
        engine: sa_engine.Engine,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        uri: typing.Optional[str] = None,
        engine: typing.Optional[sa_engine.Engine] = None,
    ) -> None:
        if not any((uri, engine)):
            raise TypeError("Missing parameters (uri/engine)")
        self._uri = uri
        if engine is not None:
            self._engine = engine

    @lazyfield
    def _engine(self):
        assert self._uri
        return sa.create_engine(self._uri)

    def is_closed(self, client: sa_engine.Connection) -> bool:
        return client.closed

    def new(self):
        return self._engine.connect()

    def release(self, client: sa_engine.Connection) -> None:
        client.close()

    def begin(self, client: sa_engine.Connection) -> None:
        if client.in_transaction():
            client.begin_nested()
        else:
            client.begin()

    def commit(self, client: sa_engine.Connection) -> None:
        if trx := (client.get_nested_transaction() or client.get_transaction()):
            trx.commit()

    def rollback(self, client: sa_engine.Connection) -> None:
        if trx := (client.get_nested_transaction() or client.get_transaction()):
            trx.rollback()

    def in_atomic(self, client: sa_engine.Connection) -> bool:
        return client.in_transaction()

    def context(
        self,
    ) -> "SaContext":
        return SaContext(self)


class SessionAdapter(context.AtomicAdapter[sa_orm.Session]):
    def __init__(self, adapter: context.AtomicAdapter[sa_engine.Connection]) -> None:
        self._internal_adapter = adapter

    def new(self):
        return sa_orm.Session(self._internal_adapter.new())

    def release(self, client: sa_orm.Session) -> None:
        self._internal_adapter.release(client.get_bind())  # type: ignore

    def is_closed(self, client: sa_orm.Session) -> bool:
        # sqlalchemy session is never truly closed
        return self._internal_adapter.is_closed(client.get_bind())  # type: ignore

    def begin(self, client: sa_orm.Session) -> None:
        if client.in_transaction():
            client.begin_nested()
        else:
            client.begin()

    def commit(self, client: sa_orm.Session) -> None:
        if trx := (client.get_nested_transaction() or client.get_transaction()):
            trx.commit()

    def rollback(self, client: sa_orm.Session) -> None:
        if trx := (client.get_nested_transaction() or client.get_transaction()):
            trx.rollback()

    def in_atomic(self, client: sa_orm.Session) -> bool:
        return client.in_transaction()

    def context(
        self,
    ) -> "SessionContext":
        return SessionContext(self)



import re
import shlex
from typing import Callable
from typing import TypeVar

_to_camel_regexp = re.compile("_([a-zA-Z])")
_to_snake_regexp = re.compile("([a-z])([A-Z])")


def to_snake(camel_string: str) -> str:
    """
    The to_snake function converts camelCase strings to snake_case.

    :param camel_string: str: Pass the camelcase string to be converted
    :return: A string in snake_case
    """
    return _to_snake_regexp.sub(r"\1_\2", camel_string).lower()


def to_camel(snake_string: str) -> str:
    """
    The to_camel function converts a snake_string to camelCase.

    :param snake_string: str: Pass the string that is to be converted
    :return: A string that is converted from snake_case to camelcase
    """
    return _to_camel_regexp.sub(
        lambda match: match[1].upper(), snake_string
    ).removesuffix("_")


def upper_camel(snake_string: str):
    """
    The upper_camel function converts a snake_case string to an upperCamelCase string.

    :param snake_string: str: Pass in the string that we want to convert
    :return: A string that is the snake_string converted to camel
    case and with the first letter capitalized
    """
    val = to_camel(snake_string)
    return val and val[0].upper() + val[1:]


T = TypeVar("T")

OuterCastT = TypeVar("OuterCastT", list, tuple, set)


def make_lex_separator(
    outer_cast: type[OuterCastT], cast: type = str
) -> Callable[[str], OuterCastT]:
    def wrapper(value: str) -> OuterCastT:
        lex = shlex.shlex(value, posix=True)
        lex.whitespace = ","
        lex.whitespace_split = True
        return outer_cast(cast(item.strip()) for item in lex)

    return wrapper


comma_separated = make_lex_separator(tuple, str)
